from tweepy import Client
from ..api import SourceBase, Query, ANDQuery, CoverageEntry


class TwitterSource(SourceBase):
    def __init__(self,  access_token, access_token_secret,
                 consumer_key, consumer_secret, bearer_token):
        self._client = Client(
            access_token=access_token,
            access_token_secret=access_token_secret,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            bearer_token=bearer_token
        )

    def convert_query(self, query: Query) -> str:
        if isinstance(query, ANDQuery):
            return "-is:retweet " + " ".join(query.keywords)
        else:
            raise TypeError("Only supporting AND queries at the moment") 

    def get_query_results(self, query: str, from_timestamp: int = None):
        posts = []
        kwargs = {
            "query": query,
            "expansions": ['author_id', 'attachments.media_keys'],
            "user_fields": ['name', 'username', 'profile_image_url'],
            "tweet_fields": ['created_at', 'context_annotations'],
            "media_fields": ['preview_image_url', 'height', 'url']
        }
        new_posts, incs, x, info = self._client.search_recent_tweets(
            start_time=from_timestamp,
            **kwargs
        )

        while new_posts:
            if 'media' not in incs:
                incs['media'] = [None for _ in range(len(new_posts))]
            posts.extend([
                (post, inc, media)
                for post, inc, media
                in zip(new_posts, incs['users'], incs['media'])
                ]
            )
            new_posts, incs, _, info = self._client.search_recent_tweets(
                since_id=info['newest_id'],
                **kwargs
            )
        return posts
    
    def result_to_entry(self, result) -> CoverageEntry:
        entry = self._create_new_entry(
            actor_primary=result[1].name,
            actor_secondary=result[1].username,
            actor_logo=result[1].profile_image_url,
            date=result[0].created_at,
            body=result[0].text,
            title=result[0].text,
            image_url=getattr(result[2], 'url', None)
        )
        return entry



"""
"""
import favicon
from datetime import datetime

from newsapi import NewsApiClient
from newsapi.newsapi_exception import NewsAPIException
from ..api import SourceBase, CoverageEntry, Query, ANDQuery


class NewsAPISource(SourceBase):
    PAGE_SIZE = 100

    def __init__(self, api_key):
        self._client = NewsApiClient(api_key=api_key)

    def convert_query(self, query: Query) -> str:
        if isinstance(query, ANDQuery):
            return " AND ".join(query.keywords)
        else:
            raise TypeError("Only supporting AND queries at the moment")

    def get_query_results(self, query: str, from_timestamp: datetime = None):
        try:
            response = self._client.get_everything(
                q=query,
                from_param=from_timestamp,
                sort_by='publishedAt',
                page_size=self.PAGE_SIZE
            )
        except NewsAPIException as e:
            response = self._client.get_everything(
                q=query,
                sort_by='publishedAt',
                page_size=self.PAGE_SIZE
            )
        return response['articles']

    def result_to_entry(self, result) -> CoverageEntry:
        logo = None
        icons = favicon.get(result['url'])
        if len(icons) > 0:
            logo = icons[0]
        entry = self._create_new_entry(
            actor_primary=result['source']['name'],
            actor_secondary=result['author'],
            actor_logo=logo.url,
            date=datetime.strptime(
                result['publishedAt'],
                "%Y-%m-%dT%H:%M:%SZ"
            ),
            body=result['description'],
            title=result['title'],
            url=result['url'],
            image_url=result['urlToImage']
        )
        return entry



from datetime import datetime
from ..api import SourceBase, CoverageEntry, Query, ANDQuery

from praw import Reddit


class RedditSource(SourceBase):
    def __init__(self, client_id, client_secret, user_agent):
        self._client = Reddit(
            client_id=client_id,
            client_secret=client_secret,
            user_agent=user_agent
        )

    def convert_query(self, query: Query) -> str:
        if isinstance(query, ANDQuery):
            return " AND ".join(
                [
                    f"((self:yes selftext:{kwrd}) OR (title:{kwrd}))"
                    for kwrd in query.keywords
                ]
            )
        else:
            raise TypeError("Only supporting AND queries at the moment")

    def get_query_results(self, query: str, from_timestamp: datetime = None):
        listings = self._client.subreddit("all").search(query, sort='new')
        if from_timestamp is None:
            return listings
        else:
            res = []
            for post in listings:
                if post.created_utc > from_timestamp.timestamp():
                    res.append(post)
                else:
                    break
            return res

    def result_to_entry(self, result) -> CoverageEntry:
        entry = self._create_new_entry(
            actor_primary=result.author.name,
            actor_secondary=result.subreddit.display_name,
            score=result.score,
            actor_logo=getattr(result.author, "icon_img", None),
            date=datetime.fromtimestamp(result.created_utc),
            body=result.selftext,
            title=result.title,
            url=result.url
        )
        return entry



"""Jekyll aggregator.

Fetches entries from sources and converts them into frontmatter Posts.
"""
from typing import Dict, Any

import hashlib
from datetime import datetime
import frontmatter
from frontmatter import Post
from dataclasses import dataclass, asdict
import json
import os

from ..api import CoverageAggreagatorBase, CoverageEntry


@dataclass
class JekyllCoverageAggregator(CoverageAggreagatorBase):
    """Jekyll coverage aggregator.

    Attributes:
        post_location: Output path for frontmatter posts.
        last_post_dates_file_name: Path to file storing dates of last posts
            for each source.
    """
    post_location: str
    last_post_dates_file_name: str 

    def get_last_entry_date(self, source_name: str) -> datetime:
        """Gets date of the last entry for the source.

        Args:
            source_name: Name of the source to fetch the date for.

        Returns:
            A datetime object representing the date of the last entry
            in UTC.
        """
        if not os.path.exists(self.last_post_dates_file_name):
            with open(self.last_post_dates_file_name, 'w') as f:
                json.dump(
                    {s.__class__.__name__: None for s in self.sources},
                    f
                )

        with open(self.last_post_dates_file_name, 'r') as f:
            sources = json.load(f)
            timestamp = sources.get(source_name, None)
            if timestamp is not None:
                return datetime.fromtimestamp(int(timestamp)+1)
            else:
                return None

    def set_last_entry_date(
            self, source_name: str, timestamp: datetime) -> None:
        """Sets date of the last entry for the source.

        Args:
            source_name: Name of the source to set the date for.
            timestamp: Timestamp to set.
        """
        with open(self.last_post_dates_file_name, 'r') as f:
            sources = json.load(f)

        sources[source_name] = timestamp.timestamp()

        with open(self.last_post_dates_file_name, 'w') as f:
            f.write(json.dumps(sources))

    @staticmethod
    def entry_to_tags(entry: CoverageEntry) -> Dict[str, Any]:
        """Converts entry to dictionary of post tags.

        Args:
            entry: Input entry.

        Returns:
            A dictionary of tags.
        """
        tags = asdict(entry)
        del(tags['body'])
        tags['date'] = tags['date'].timestamp()
        return tags

    @staticmethod
    def encode_tags(tags: Dict[str, Any]) -> str:
        """Encodes tags to md5 hash.

        Helps to assign a unique id for each post.

        Args:
            tags: A dictionary of tags.

        Returns:
            md5 hexidigest string.
        """
        m = hashlib.md5()
        keys = ['actor_primary', 'actor_secondary', 'date', '_source_cls']
        for key in keys:
            m.update(str(tags.get(key)).encode())
        return m.hexdigest()

    def save_entry(self, entry: CoverageEntry) -> None:
        """Saves entry to the post location.
        
        Converts entry to a post, and saves it to the post location.

        Args:
            entry: Input entry.
        """
        tags = self.entry_to_tags(entry)
        entry_timestamp = tags['date']

        created_at = datetime.fromtimestamp(entry_timestamp).strftime(
            "%Y-%m-%d"
        )
        title = self.encode_tags(tags)
        new_entry_file_name = f"{created_at}-{title}.txt"

        new_entry_path = os.path.join(self.post_location, new_entry_file_name)

        if not os.path.exists(new_entry_path):
            with open(new_entry_path, 'wb') as f:
                post = Post(
                    content=entry.body,
                    osma=tags
                )
                frontmatter.dump(post, fd=f)



"""OSMA CLI
"""

import sys
import click
import toml
from loguru import logger

from osma.api import SourceMeta, AggregatorMeta, QueryMeta

logger.add(
    sys.stderr,
    format="{time} {level} {message}",
    level="INFO"
)


@click.group()
@click.option('--config', '-c', type=click.File('r'))
@click.pass_context
def osma(ctx, config):
    conf_dict = toml.load(config)
    ctx.ensure_object(dict)

    ctx.obj['sources'] = []
    if 'sources' in conf_dict:
        for _, source_dict in conf_dict['sources'].items():
            source_type_name = source_dict.get('type', None)
            if source_type_name in SourceMeta.__sources__:
                del source_dict['type']
                source = SourceMeta.__sources__[source_type_name](
                    **source_dict
                )
                ctx.obj['sources'].append(source)

    ctx.obj['aggregators'] = []

    if 'aggregators' in conf_dict:
        for _, agg_dict in conf_dict['aggregators'].items():
            agg_type_name = agg_dict.get('type', None)
            if agg_type_name in AggregatorMeta.__aggs__:
                del agg_dict['type']

                query = None
                if 'query' in agg_dict:
                    query_dict = agg_dict.get('query', {})
                    query_type_name = query_dict.get('type', None)
                    if query_type_name in QueryMeta.__queries__:
                        del query_dict['type']
                        query = QueryMeta.__queries__[query_type_name](
                           ** query_dict
                        )
                del agg_dict['query']
                agg = AggregatorMeta.__aggs__[agg_type_name](
                    sources=ctx.obj['sources'],
                    query=query,
                    **agg_dict
                )
                ctx.obj['aggregators'].append(agg)


@osma.command()
@click.pass_context
def run(ctx):
    for aggregator in ctx.obj['aggregators']:
        logger.info(f'Starting {aggregator.__class__.__name__}')
        aggregator.run()


if __name__ == '__main__':
    osma()



from .api import (
    CoverageEntry,
    Query, ANDQuery, ORQuery,
    SourceBase, CoverageAggreagatorBase
)
from .sources import RedditSource, TwitterSource, NewsAPISource
from .aggregators import JekyllCoverageAggregator
from .cli import osma

__all__ = [
    "CoverageEntry",
    "Query", "ANDQuery", "ORQuery",
    "SourceBase", "CoverageAggreagatorBase",
    "RedditSource", "TwitterSource", "NewsAPISource",
    "JekyllCoverageAggregator",
    "osma"
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['osma', 'osma.aggregators', 'osma.sources']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'favicon>=0.7.0,<0.8.0',
 'loguru>=0.5.3,<0.6.0',
 'newsapi-python>=0.2.6,<0.3.0',
 'praw>=7.5.0,<8.0.0',
 'python-frontmatter>=1.0.0,<2.0.0',
 'toml>=0.10.2,<0.11.0',
 'tweepy>=4.4.0,<5.0.0']

entry_points = \
{'console_scripts': ['osma = osma.cli:osma']}

setup_kwargs = {
    'name': 'osma',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Mihails Delmans',
    'author_email': 'm.delmans@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""OSMA core API.
"""

from typing import List, TypeVar, Optional
from datetime import datetime
from dataclasses import dataclass
from abc import ABCMeta, abstractmethod
from loguru import logger


@dataclass
class CoverageEntry:
    """Coverage entry.
    
    A standard data class for representing a coverage entry.
    
    Args:
        actor_primary: A primary name of the author.
        actor_secondary: A secondary name of the author.
        score: Number representing engagement with the entry, for e.g.
            number of likes or upvotes.
        country: Country of entry origin.
        actor_logo: URL to the actor's logo.
        date: Date of the entry creation.
        body: Body of the entry.
        url: URL to the full entry.
        title: Title of the entry.
        image_url: URL to the media image.

    Attrs:
        actor_primary: A primary name of the author.
        actor_secondary: A secondary name of the author.
        score: Number representing engagement with the entry, for e.g.
            number of likes or upvotes.
        country: Country of entry origin.
        actor_logo: URL to the actor's logo.
        date: Date of the entry creation.
        body: Body of the entry.
        url: URL to the full entry.
        title: Title of the entry.
        image_url: URL to the media image.
    """
    _source_cls: str
    actor_primary: str
    actor_secondary: str
    date: datetime
    body: str
    score: Optional[int] = None
    country: Optional[str] = None
    actor_logo: Optional[str] = None
    url: Optional[str] = None
    title: Optional[str] = None
    image_url: Optional[str] = None


class QueryMeta(type):
    __queries__ = {}

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if name not in QueryMeta.__queries__:
            QueryMeta.__queries__[name] = cls


class Query(metaclass=QueryMeta):
    """A base class for defying queries"""
    pass


@dataclass
class ANDQuery(Query):
    """AND query.

    A query that is formed by matching all given keywords.

    Attrs:
        keywords: A list of keywords.
    """
    keywords: List[str]


@dataclass
class ORQuery(Query):
    """OR query.

    A query that is formed by matching at least one of the given
    keywords.

    Attrs:
        keywords: A list of keywords.
    """
    keywords: List[str]


R = TypeVar('R')


class SourceMeta(ABCMeta):
    __sources__ = {}

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if name not in SourceMeta.__sources__:
            SourceMeta.__sources__[name] = cls


class SourceBase(metaclass=SourceMeta):
    """A source base class.

    Source classes are responsible for fetching new entries for the given
    query and converting them to coverage entries.
    """
    @abstractmethod
    def convert_query(self, query: Query) -> str:
        """Converts query from a standard definition to a string
        specific to the source.

        Args:
            query: A query object.

        Returns:
            A string representation of the source-specific query.

        Raises:
            TypeError: If query type is not supported.
        """
        pass

    @abstractmethod
    def get_query_results(
            self,
            query: str,
            from_timestamp: Optional[datetime] = None
            ) -> List[R]:
        """Gets intermediate query result.

        Args:
            query: A string representing source-specific query.
            from_timestamp: Minimal datetime of the entry to querry.

        Returns:
            A list of objects, each rrepresenting an enrty in a
            source-specific format. If ``from_timestamp`` is given,
            must only return entries that are past the timestamp.
        """
        pass

    @abstractmethod
    def result_to_entry(self, result: R) -> CoverageEntry:
        """Converts entry from a soure-specific format to a standard
        entry object.

        Args:
            result: A source-specific entry object.

        Returns:
            Converted coverage entry object. 
        """
        pass

    def fetch_entries(
            self,
            query: Query,
            from_timestamp: datetime = None
            ) -> List[CoverageEntry]:
        """Fetches entries from the source.

        Args:
            query: A query to use for fetching the entries.
            from_timestamp: A minimal date of entry.

        Returns:
            A list of the coverage entries that satisfy the given
            ``query`` and are past the given timestamp.
        """
        try:
            specific_query = self.convert_query(query)
        except BaseException as e:
            raise RuntimeError(
                "Could not convert input query to a specific one"
            ) from e
        try:
            results = self.get_query_results(specific_query, from_timestamp)
        except ConnectionError as e:
            raise ConnectionError("Failed to get query results") from e
        except BaseException as e:
            raise RuntimeError("Failed to get query results") from e

        for res in results:
            try:
                yield self.result_to_entry(res)
            except BaseException as e:
                print(f"Could not convert result to entry: {e}")

    @classmethod
    def _create_new_entry(cls, **kwargs):
        return CoverageEntry(
            _source_cls=cls.__name__,
            **kwargs
        )


class AggregatorMeta(ABCMeta):
    __aggs__ = {}

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if name not in AggregatorMeta.__aggs__:
            AggregatorMeta.__aggs__[name] = cls


@dataclass
class CoverageAggreagatorBase(metaclass=AggregatorMeta):
    """A base class for coverage aggregator.
    
    Aggregators collect coverage entries from the given set of sources
    using the given standard query and save them.
    
    Args:
        sources: A list of sources.
        query: A query to use for fetching the entries.

    Attrs:
        sources: A list of sources.
        query: A query to use for fetching the entries.
    """
    sources: List[SourceBase]
    query: str

    @abstractmethod
    def save_entry(self, entry: CoverageEntry) -> None:
        """Saves entry.
        """
        pass

    @abstractmethod
    def get_last_entry_date(self, source_name: str) -> Optional[datetime]:
        """Gets the date of the last entry for the given source.

        The returned timestamp is used as a ``from_timestamp`` argument
        for filtering query results in the sources.

        Should be used in conjunction with ``set_last_entry_date`` to
        create a system for saving the dates of the last entries for
        every source.

        Args:
            source_name: Name of the source class to get the last entry
            date for.
        Returns:
            Date of the last fetched entry.
        """
        pass

    @abstractmethod
    def set_last_entry_date(self, source_name: str,
                            timestamp: datetime) -> None:
        """Sets the date of the last entry for the given source.

        Should be used in conjunction with ``set_last_entry_date`` to
        create a system for saving the dates of the last entries for
        every source.

        Args:
            source_name: Name of the source class to set the last entry
            date for.
            timestamp: Date to set.
        """
        pass

    def run(self) -> None:
        """Runs aggregator.

        For each source, fetches entries that are newer than the last
        entry date as defined in ``get_last_entry_date`` and saves them
        using ``save_entry`` method. Then updates the last entry date
        using ``set_last_entry_date`` method.
        """
        for source in self.sources:
            source_name = source.__class__.__name__
            logger.info(f'Collecting data from {source_name}...')
            last_entry_date = self.get_last_entry_date(source_name)

            new_entries = source.fetch_entries(
                self.query,
                from_timestamp=last_entry_date
            )

            for entry in new_entries:
                logger.info(f'Fetched new entry {entry.title}')
                self.save_entry(entry)
                if (
                        last_entry_date is None
                        or entry.date > last_entry_date
                ):
                    last_entry_date = entry.date
            
            if last_entry_date is not None:
                self.set_last_entry_date(source_name, last_entry_date)
